@@ -2,7 +2,7 @@
     This file is part of the MTP KIOD module, part of the KDE project.
 
     SPDX-FileCopyrightText: 2018 Andreas Krutzler <andreas.krutzler@gmx.net>
-    SPDX-FileCopyrightText: 2022 Andreas Krutzler <andreas.krutzler@gmx.net>
+    SPDX-FileCopyrightText: 2022-2023 Harald Sitter <sitter@kde.org>
 
     SPDX-License-Identifier: GPL-2.0-or-later
 */
@@ -10,25 +10,35 @@
 #include "mtpstorage.h"
 
 #include <memory>
+#include <span>
 
 #include <QDateTime>
 #include <qplatformdefs.h>
 
+#include "config-mtp.h"
 #include "kiod_kmtpd_debug.h"
+#include "memory.h"
 #include "mtpdevice.h"
+#include "mtpfile.h"
+#include "mtplister.h"
 #include "storageadaptor.h"
 
-namespace std
+#if defined(HAVE_LIBMTP_Get_Children)
+static QString dbusErrorNOENT()
 {
-// Add a default deleter for LIBMTP_file_t. Means we do not have to repeat the deleter all over the place.
-template<>
-struct default_delete<LIBMTP_file_t> {
-    void operator()(LIBMTP_file_t *ptr) const
-    {
-        LIBMTP_destroy_file_t(ptr);
-    }
-};
-} // namespace std
+    return QStringLiteral("org.kde.kmtp.Error.NoEntry");
+}
+
+static QString dbusErrorNOTDIR()
+{
+    return QStringLiteral("org.kde.kmtp.Error.NotDirectory");
+}
+#else
+static QString dbusErrorENOSYS()
+{
+    return QStringLiteral("org.kde.kmtp.Error.NotImplemented");
+}
+#endif
 
 /**
  * @brief MTPDataPutFunc callback function, "puts" data from the device somewhere else
@@ -161,121 +171,6 @@ static LIBMTP_filetype_t getFiletype(const QString &filename)
     return filetype;
 }
 
-/**
- * @brief Converts a LIBMTP_filetype_t to a mime-type
- */
-static QString getMimetype(LIBMTP_filetype_t filetype)
-{
-    switch (filetype) {
-    case LIBMTP_FILETYPE_FOLDER:
-        return QStringLiteral("inode/directory");
-    case LIBMTP_FILETYPE_WAV:
-        return QStringLiteral("audio/wav");
-    case LIBMTP_FILETYPE_MP3:
-        return QStringLiteral("audio/x-mp3");
-    case LIBMTP_FILETYPE_WMA:
-        return QStringLiteral("audio/x-ms-wma");
-    case LIBMTP_FILETYPE_OGG:
-        return QStringLiteral("audio/x-vorbis+ogg");
-    case LIBMTP_FILETYPE_AUDIBLE:
-        return {};
-    case LIBMTP_FILETYPE_MP4:
-        return QStringLiteral("video/mp4");
-    case LIBMTP_FILETYPE_UNDEF_AUDIO:
-        return {};
-    case LIBMTP_FILETYPE_WMV:
-        return QStringLiteral("video/x-ms-wmv");
-    case LIBMTP_FILETYPE_AVI:
-        return QStringLiteral("video/x-msvideo");
-    case LIBMTP_FILETYPE_MPEG:
-        return QStringLiteral("video/mpeg");
-    case LIBMTP_FILETYPE_ASF:
-        return QStringLiteral("video/x-ms-asf");
-    case LIBMTP_FILETYPE_QT:
-        return QStringLiteral("video/quicktime");
-    case LIBMTP_FILETYPE_UNDEF_VIDEO:
-        return {};
-    case LIBMTP_FILETYPE_JPEG:
-        return QStringLiteral("image/jpeg");
-    case LIBMTP_FILETYPE_JFIF:
-        return {};
-    case LIBMTP_FILETYPE_TIFF:
-        return QStringLiteral("image/tiff");
-    case LIBMTP_FILETYPE_BMP:
-        return QStringLiteral("image/bmp");
-    case LIBMTP_FILETYPE_GIF:
-        return QStringLiteral("image/gif");
-    case LIBMTP_FILETYPE_PICT:
-        return QStringLiteral("image/x-pict");
-    case LIBMTP_FILETYPE_PNG:
-        return QStringLiteral("image/png");
-    case LIBMTP_FILETYPE_VCALENDAR1:
-        return QStringLiteral("text/x-vcalendar");
-    case LIBMTP_FILETYPE_VCALENDAR2:
-        return QStringLiteral("text/x-vcalendar");
-    case LIBMTP_FILETYPE_VCARD2:
-        return QStringLiteral("text/x-vcard");
-    case LIBMTP_FILETYPE_VCARD3:
-        return QStringLiteral("text/x-vcard");
-    case LIBMTP_FILETYPE_WINDOWSIMAGEFORMAT:
-        return QStringLiteral("image/x-wmf");
-    case LIBMTP_FILETYPE_WINEXEC:
-        return QStringLiteral("application/x-ms-dos-executable");
-    case LIBMTP_FILETYPE_TEXT:
-        return QStringLiteral("text/plain");
-    case LIBMTP_FILETYPE_HTML:
-        return QStringLiteral("text/html");
-    case LIBMTP_FILETYPE_FIRMWARE:
-        return {};
-    case LIBMTP_FILETYPE_AAC:
-        return QStringLiteral("audio/aac");
-    case LIBMTP_FILETYPE_MEDIACARD:
-        return {};
-    case LIBMTP_FILETYPE_FLAC:
-        return QStringLiteral("audio/flac");
-    case LIBMTP_FILETYPE_MP2:
-        return QStringLiteral("video/mpeg");
-    case LIBMTP_FILETYPE_M4A:
-        return QStringLiteral("audio/mp4");
-    case LIBMTP_FILETYPE_DOC:
-        return QStringLiteral("application/msword");
-    case LIBMTP_FILETYPE_XML:
-        return QStringLiteral("text/xml");
-    case LIBMTP_FILETYPE_XLS:
-        return QStringLiteral("application/vnd.ms-excel");
-    case LIBMTP_FILETYPE_PPT:
-        return QStringLiteral("application/vnd.ms-powerpoint");
-    case LIBMTP_FILETYPE_MHT:
-        return {};
-    case LIBMTP_FILETYPE_JP2:
-        return QStringLiteral("image/jpeg2000");
-    case LIBMTP_FILETYPE_JPX:
-        return QStringLiteral("application/x-jbuilder-project");
-    case LIBMTP_FILETYPE_ALBUM:
-    case LIBMTP_FILETYPE_PLAYLIST:
-    case LIBMTP_FILETYPE_UNKNOWN:
-        return {};
-    }
-    return {};
-}
-
-/**
- * @brief Creates @ref KMTPFile from LIBMTP_file_t
- *
- * @param file  Must not be a nullptr!
- */
-static KMTPFile createMTPFile(const std::unique_ptr<LIBMTP_file_t> &file)
-{
-    Q_ASSERT(file.get());
-    return KMTPFile(file->item_id,
-                    file->parent_id,
-                    file->storage_id,
-                    file->filename,
-                    file->filesize,
-                    file->modificationdate,
-                    getMimetype(file->filetype));
-}
-
 MTPStorage::MTPStorage(const QString &dbusObjectPath, const LIBMTP_devicestorage_t *mtpStorage, MTPDevice *parent)
     : QObject(parent),
       m_dbusObjectPath(dbusObjectPath)
@@ -344,37 +239,38 @@ KMTPFile MTPStorage::getFileFromPath(const QString &path)
     if (!pathItems.isEmpty()) {
 
         // 1. check if the file is in the cache
-        const quint32 itemId = queryPath(path);
-        if (itemId) {
+        const auto itemId = queryPath(path);
+        if (itemId.has_value()) {
             qCDebug(LOG_KIOD_KMTPD) << "Match found in cache, checking device";
 
-            std::unique_ptr<LIBMTP_file_t> file(LIBMTP_Get_Filemetadata(getDevice(), itemId));
+            std::unique_ptr<LIBMTP_file_t> file(LIBMTP_Get_Filemetadata(getDevice(), itemId.value()));
             if (file) {
                 qCDebug(LOG_KIOD_KMTPD) << "Found file in cache";
-                const KMTPFile mtpFile = createMTPFile(file);
-                return mtpFile;
+                return createKMTPFile(file);
             }
         }
 
         // 2. query cache for parent
         else if (pathItems.size() > 1) {
             QString parentPath = convertToPath(pathItems, pathItems.size() - 1);
-            quint32 parentId = queryPath(parentPath);
+            const auto parentId = queryPath(parentPath);
 
-            qCDebug(LOG_KIOD_KMTPD)  << "Match for parent found in cache, checking device. Parent id = " << parentId;
+            if (parentId.has_value()) {
+                qCDebug(LOG_KIOD_KMTPD)  << "Match for parent found in cache, checking device. Parent id = " << parentId.value();
 
-            std::unique_ptr<LIBMTP_file_t> parent(LIBMTP_Get_Filemetadata(getDevice(), parentId));
-            if (parent) {
-                qCDebug(LOG_KIOD_KMTPD) << "Found parent in cache";
+                std::unique_ptr<LIBMTP_file_t> parent(LIBMTP_Get_Filemetadata(getDevice(), parentId.value()));
+                if (parent) {
+                    qCDebug(LOG_KIOD_KMTPD) << "Found parent in cache";
 
-                const KMTPFileList list = getFilesAndFoldersCached(parentPath, parentId);
-                const auto it = std::find_if(list.constBegin(), list.constEnd(), [pathItems](const KMTPFile &file) {
-                    return file.filename() == pathItems.last();
-                });
+                    const KMTPFileList list = getFilesAndFoldersCached(parentPath, parentId.value());
+                    const auto it = std::find_if(list.constBegin(), list.constEnd(), [pathItems](const KMTPFile &file) {
+                        return file.filename() == pathItems.last();
+                    });
 
-                if (it != list.constEnd()) {
-                    qCDebug(LOG_KIOD_KMTPD) << "Found file from cached parent";
-                    return *it;
+                    if (it != list.constEnd()) {
+                        qCDebug(LOG_KIOD_KMTPD) << "Found file from cached parent";
+                        return *it;
+                    }
                 }
             }
         }
@@ -385,17 +281,13 @@ KMTPFile MTPStorage::getFileFromPath(const QString &path)
     quint32 currentParent = LIBMTP_FILES_AND_FOLDERS_ROOT;
 
     for (const QString &element : pathItems) {
-        const KMTPFileList list = getFilesAndFoldersCached(currentPath, currentParent);
-        auto it = std::find_if(list.constBegin(), list.constEnd(), [element] (const KMTPFile &file) {
-            return file.filename() == element;
-        });
-
-        if (it != list.constEnd()) {
-            currentParent = it->itemId();
-        } else {
+        std::optional<KMTPFile> optionalFile = findEntry(element, currentPath, currentParent);
+        if (!optionalFile.has_value()) {
             qCDebug(LOG_KIOD_KMTPD) << "File not found!";
             return {};
         }
+
+        currentParent = optionalFile->itemId();
         currentPath.append(QLatin1Char('/') + element);
     }
 
@@ -403,8 +295,7 @@ KMTPFile MTPStorage::getFileFromPath(const QString &path)
     if (file) {
         qCDebug(LOG_KIOD_KMTPD) << "Found file using tree walk";
 
-        const KMTPFile mtpFile = createMTPFile(file);
-        return mtpFile;
+        return createKMTPFile(file);
     }
     return {};
 }
@@ -415,7 +306,7 @@ KMTPFileList MTPStorage::getFilesAndFoldersCached(const QString &path, quint32 p
 
     std::unique_ptr<LIBMTP_file_t> file(LIBMTP_Get_Files_And_Folders(getDevice(), m_id, parentId));
     while (file != nullptr) {
-        const KMTPFile mtpFile = createMTPFile(file);
+        const KMTPFile mtpFile = createKMTPFile(file);
         addPath(path + QLatin1Char('/') + mtpFile.filename(), mtpFile.itemId());
         mtpFiles.append(mtpFile);
 
@@ -424,7 +315,45 @@ KMTPFileList MTPStorage::getFilesAndFoldersCached(const QString &path, quint32 p
     return mtpFiles;
 }
 
-quint32 MTPStorage::queryPath(const QString &path, int timeToLive)
+#if defined(HAVE_LIBMTP_Get_Children)
+std::optional<KMTPFile> MTPStorage::findEntry(const QString &fileNeedle, const QString &parentPath, quint32 parentId)
+{
+    // Optimized version of finding. Using LIBMTP_Get_Children to get a list of all ids relatively quickly and then
+    // iterate them for the needle - leading to an early return when the needle is found, and partial caching!
+    uint32_t *children = nullptr;
+    const auto childrenSize = LIBMTP_Get_Children(getDevice(), m_id, parentId, &children);
+    if (childrenSize == -1) {
+        return std::nullopt;
+    }
+    const std::unique_ptr<uint32_t> childrenOwner(children);
+
+    for (const auto &child : std::span<uint32_t>(children, childrenSize)) {
+        const KMTPFile mtpFile = createKMTPFile(std::unique_ptr<LIBMTP_file_t>(LIBMTP_Get_Filemetadata(getDevice(), child)));
+        addPath(parentPath + QLatin1Char('/') + mtpFile.filename(), mtpFile.itemId());
+        if (mtpFile.filename() == fileNeedle) {
+            return mtpFile;
+        }
+    }
+
+    return std::nullopt;
+}
+#else
+std::optional<KMTPFile> MTPStorage::findEntry(const QString &fileNeedle, const QString &parentPath, quint32 parentId)
+{
+    // Poor man's search function. This gets all metadata in advance meaning even if the needle is the first of N
+    // entries we'll have gotten all N entries.
+    const KMTPFileList list = getFilesAndFoldersCached(parentPath, parentId);
+    auto it = std::find_if(list.constBegin(), list.constEnd(), [fileNeedle](const KMTPFile &file) {
+        return file.filename() == fileNeedle;
+    });
+    if (it == list.cend()) {
+        return std::nullopt;
+    }
+    return *it;
+}
+#endif
+
+std::optional<quint32> MTPStorage::queryPath(const QString &path, int timeToLive)
 {
     QPair< QDateTime, uint32_t > item = m_cache.value(path);
 
@@ -437,10 +366,10 @@ quint32 MTPStorage::queryPath(const QString &path, int timeToLive)
             return item.second;
         }
         m_cache.remove(path);
-        return 0;
+        return std::nullopt;
     }
 
-    return 0;
+    return std::nullopt;
 }
 
 void MTPStorage::addPath(const QString &path, quint32 id, int timeToLive)
@@ -550,7 +479,7 @@ int MTPStorage::sendFileFromFileDescriptor(const QDBusUnixFileDescriptor &descri
 
     QTimer::singleShot(0, this, [this, parentId, descriptor, filename] {
         int result = 1;
-        QT_STATBUF srcBuf;
+        QT_STATBUF srcBuf{};
         if (QT_FSTAT(descriptor.fileDescriptor(), &srcBuf) != -1) {
             const QDateTime lastModified = QDateTime::fromSecsSinceEpoch(srcBuf.st_mtim.tv_sec);
 
@@ -559,9 +488,10 @@ int MTPStorage::sendFileFromFileDescriptor(const QDBusUnixFileDescriptor &descri
             file->filename = qstrdup(filename.toUtf8().data());
             file->filetype = getFiletype(filename);
             file->filesize = quint64(srcBuf.st_size);
-            file->modificationdate = lastModified.toSecsSinceEpoch();   // no matter what to set here, current time is taken
+            file->modificationdate = lastModified.toSecsSinceEpoch(); // no matter what to set here, current time is taken
             file->storage_id = m_id;
 
+            QT_LSEEK(descriptor.fileDescriptor(), 0, SEEK_SET); // make sure we are at the beginning of the file, the worker may leave it sitting at the end
             result = LIBMTP_Send_File_From_File_Descriptor(getDevice(), descriptor.fileDescriptor(), file.get(), onDataProgress, this);
 
             if (result) {
@@ -580,17 +510,20 @@ int MTPStorage::setFileName(const QString &path, const QString &newName)
     qCDebug(LOG_KIOD_KMTPD) << "setFileName:" << path << newName;
 
     const KMTPFile file = getFileFromPath(path);
-    if (file.isValid()) {
-        std::unique_ptr<LIBMTP_file_t> source(LIBMTP_Get_Filemetadata(getDevice(), file.itemId()));
-        if (source) {
-            const int result = LIBMTP_Set_File_Name(getDevice(), source.get(), newName.toUtf8().constData());
-            if (!result) {
-                removePath(path);
-            }
-            return result;
-        }
+    if (!file.isValid()) {
+        return 1;
     }
-    return 1;
+
+    std::unique_ptr<LIBMTP_file_t> source(LIBMTP_Get_Filemetadata(getDevice(), file.itemId()));
+    if (!source) {
+        return 1;
+    }
+
+    const int result = LIBMTP_Set_File_Name(getDevice(), source.get(), newName.toUtf8().constData());
+    if (result == 0) {
+        removePath(path);
+    }
+    return result;
 }
 
 quint32 MTPStorage::createFolder(const QString &path)
@@ -599,9 +532,9 @@ quint32 MTPStorage::createFolder(const QString &path)
 
     quint32 folderId = 0;
     const QStringList pathItems = path.split(QLatin1Char('/'), Qt::SkipEmptyParts);
-    const quint32 destinationId = queryPath(path);
+    const auto destinationId = queryPath(path);
 
-    if (!pathItems.isEmpty() && !destinationId) {
+    if (!pathItems.isEmpty() && !destinationId.has_value()) {
         QByteArray dirName = pathItems.last().toUtf8();
 
         if (pathItems.size() == 1) {
@@ -637,4 +570,42 @@ int MTPStorage::deleteObject(const QString &path)
     return ret;
 }
 
+QDBusObjectPath MTPStorage::getFilesAndFolders2(const QString &path)
+{
+#if defined(HAVE_LIBMTP_Get_Children)
+    qint64 idToList = LIBMTP_FILES_AND_FOLDERS_ROOT;
+    if (!path.isEmpty() && path != QLatin1String("/")) {
+        const KMTPFile file = getFileFromPath(path);
+        if (!file.isValid()) {
+            sendErrorReply(dbusErrorNOENT(), path);
+            return {};
+        }
+        if (!file.isFolder()) {
+            sendErrorReply(dbusErrorNOTDIR(), path);
+            return {};
+        }
+        idToList = file.itemId();
+    }
+
+    uint32_t *children = nullptr;
+    const auto childrenSize = LIBMTP_Get_Children(getDevice(), m_id, idToList, &children);
+    if (childrenSize == -1) {
+        sendErrorReply(dbusErrorNOENT(), path);
+        return {};
+    }
+
+    auto lister = new MTPLister(std::unique_ptr<uint32_t>(children), childrenSize, getDevice(), path, this);
+    connect(lister, &MTPLister::entry, this, [this, path](const KMTPFile &file) {
+        addPath(path + QLatin1Char('/') + file.filename(), file.itemId()); // add to cache
+    });
+    static quint64 id = 0;
+    const QDBusObjectPath objectPath(QStringLiteral("/modules/kmtpd/Lister/%1").arg(id++));
+    connection().registerObject(objectPath.path(), lister);
+    return objectPath;
+#else
+    sendErrorReply(dbusErrorENOSYS(), QString::fromLatin1(Q_FUNC_INFO));
+    return {};
+#endif
+}
+
 #include "moc_mtpstorage.cpp"
