@@ -1,6 +1,6 @@
 /*
     SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-KDE-Accepted-GPL
-    SPDX-FileCopyrightText: 2020 Harald Sitter <sitter@kde.org>
+    SPDX-FileCopyrightText: 2020-2021 Harald Sitter <sitter@kde.org>
 */
 
 #include <QCoreApplication>
@@ -73,6 +73,20 @@ public:
     }
 };
 
+class SMBCPrinterDiscovery : public SMBCDiscovery
+{
+public:
+    SMBCPrinterDiscovery(const UDSEntry &entry)
+        : SMBCDiscovery(entry)
+    {
+        m_entry.fastInsert(KIO::UDSEntry::UDS_ACCESS, 0x0);
+        m_entry.fastInsert(KIO::UDSEntry::UDS_MIME_TYPE, QStringLiteral("inode/vnd.kde.kio.smb.printer"));
+        // Relative to parent, but need to set it so we can append a marker. Subsequent stat calls wouldn't know that this is a printer
+        // because libsmbc doesn't reflect that in the stat output.
+        m_entry.fastInsert(KIO::UDSEntry::UDS_URL, udsName() + QStringLiteral("?kio-printer=true"));
+    }
+};
+
 SMBCDiscovery::SMBCDiscovery(const UDSEntry &entry)
     : m_entry(entry)
       // cache the name, it may get accessed more than once
@@ -90,10 +104,10 @@ KIO::UDSEntry SMBCDiscovery::toEntry() const
     return m_entry;
 }
 
-SMBCDiscoverer::SMBCDiscoverer(const SMBUrl &url, QEventLoop *loop, SMBSlave *slave)
+SMBCDiscoverer::SMBCDiscoverer(const SMBUrl &url, QEventLoop *loop, SMBWorker *worker)
     : m_url(url)
     , m_loop(loop)
-    , m_slave(slave)
+    , m_worker(worker)
 {
 }
 
@@ -132,7 +146,7 @@ bool SMBCDiscoverer::discoverNextFileInfo()
         entry.fastInsert(KIO::UDSEntry::UDS_NAME, name);
 
         m_url.addPath(name);
-        m_slave->statToUDSEntry(m_url, st, entry); // won't produce useful error
+        m_worker->statToUDSEntry(m_url, st, entry); // won't produce useful error
         Q_EMIT newDiscovery(Discovery::Ptr(new SMBCDiscovery(entry)));
         m_url.cdUp();
         return true;
@@ -167,7 +181,9 @@ void SMBCDiscoverer::discoverNext()
         return;
     }
 
-    if (discoverNextFileInfo()) {
+    static bool printersOnly = qEnvironmentVariableIntValue("KIO_SMB_PRINTERS") > 0;
+
+    if (!printersOnly && discoverNextFileInfo()) {
         return;
     }
 
@@ -201,6 +217,13 @@ void SMBCDiscoverer::discoverNext()
         entry.fastInsert(KIO::UDSEntry::UDS_HIDDEN, 1);
     }
 
+    if (printersOnly) {
+        if (dirp->smbc_type == SMBC_PRINTER_SHARE) {
+            Q_EMIT newDiscovery(Discovery::Ptr(new SMBCPrinterDiscovery(entry)));
+        }
+        return;
+    }
+
 #if !defined(HAVE_READDIRPLUS2)
     // . and .. are always of the dir type so they are of no consequence outside
     // actual dir listing and that'd be done by readdirplus2 already
@@ -212,7 +235,7 @@ void SMBCDiscoverer::discoverNext()
     } else if (dirp->smbc_type == SMBC_FILE || dirp->smbc_type == SMBC_DIR) {
         // Set stat information
         m_url.addPath(name);
-        const int statErr = m_slave->browse_stat_path(m_url, entry);
+        const int statErr = m_worker->browse_stat_path(m_url, entry);
         if (statErr != 0) {
             // The entry can disappear in the time span between
             // listing and the stat call. There's nothing we or the user
@@ -299,3 +322,5 @@ void SMBCDiscoverer::queue()
     // SMBC is running on the loop and so the priority has no negative impact.
     QCoreApplication::postEvent(this, new QEvent(LoopEvent), Qt::LowEventPriority);
 }
+
+#include "moc_smbcdiscoverer.cpp"
